#ifndef VERSION_H
#define VERSION_H

/* remember to change both instance of the version -amol */

#ifdef NTDBG
#define LOCALSTR ",nt-rev-8.10-debug"
#else
#define LOCALSTR ",nt-rev-8.10" 
								//patches
#endif NTDBG

#endif VERSION_H
