#if 0
#include <windows.h>
STRINGTABLE DISCARDABLE
BEGIN
666 "	setenv EDITOR vi.exe"
667 "	setenv PAGER c:/bin/less"
668 "	setenv NOREBIND 1"
669 "	setenv LESS ""-B -c -h4 -i -M -q -x4 -G"""
670 "	setenv SHELL C:/bin/tcsh.exe"
671 "if ($?PROCESSOR_ARCHITECTURE) then"
672 "	setenv USER  $USERNAME"
673 "	setenv LOGNAME $USER"
674 "	setenv HOSTNAME $COMPUTERNAME"
675 "endif"
676 "umask 077"
677 "if (! $?prompt) then	# Everything else is interactive"
678 "	exit(0)"
679 "endif"
680 "unsetenv EXINIT"
681 "if ($?tcsh) then"
682 "	set autolist"
683 "	set listlinks"
684 "	set pushdtohome"
685 "	set visiblebell"
686 "	set ellipsis"
687 "	set rmstar"
688 "if ($?TERM) then"
689 "    if ($TERM == ""emacs"") then"
690 "	set prompt='%/\>'"
691 "    else"
692 "	set prompt='%{[1;34m%}%c03%{[0m%}\>'"
693 "	set prompt3='%{[1;31m%}Correct to %R ?(y|n|e)%{[0m%} '"
694 "    endif"
695 "else"
696 "	set prompt='%{[1;34m%}%c03%{[0m%}\>'"
697 "	set prompt3='%{[1;31m%}Correct to %R ?(y|n|e)%{[0m%} '"
698 "endif"
699 "	if ($?loginsh) then"
700 "		set correct=all"
701 "		set savedirs=100"
702 "		set autocorrect"
703 "	endif"
704 "	set complete=igncase"
705 "	complete which 'p/*/c/'"
706 "	complete where 'p/*/c/'"
707 "	complete start 'p/1/c/'"
708 "	complete cd 'p/1/d/'"
709 "	complete pushd 'p/1/d/'"
710 "	complete popd 'p/1/d/'"
711 "	complete bindkey 'p/*/b/'"
712 "	complete set 'p/*/v/'"
713 "	complete unset 'p/*/v/'"
714 "	bindkey -b ^W backward-delete-word"
715 "	bindkey -b ^U backward-kill-line"
716 "	bindkey -b ^R complete-word-fwd"
717 "	bindkey -b M-g list-glob"
718 "	bindkey -b M-* expand-glob"
719 "	bindkey -b M-v expand-variables"
720 "	bindkey -b M-w normalize-command"
721 "	bindkey -b M-/ e_dosify_prev"
722 "	bindkey -b M-r i-search-back"
723 "	bindkey -b M-x e_copy_to_clipboard"
724 "	bindkey -b M-y e_paste_from_clipboard"
725 "	bindkey -b N-up up-history"
726 "	bindkey -b N-down down-history"
727 "	bindkey -b N-right forward-char"
728 "	bindkey -b N-left backward-char"
729 "	bindkey -b N-del delete-char"
730 "	bindkey -b N-ins overwrite-mode"
731 "	bindkey -b N-1 which-command"
732 "	bindkey -b N-2 expand-history"
733 "	bindkey -b N-3 complete-word-raw"
734 "	bindkey -b N-home beginning-of-line"
735 "	bindkey -b N-end end-of-line"
736 "endif"
737 "if (! $?PATH_DONE) then"
738 "set path=($path:q c:/bin .)"
739 "setenv PATH_DONE  1"
740 "endif"
741 "set notify"
742 "set noclobber"
743 "set filec"
744 "set history=(200) "
745 "set savehist=(200 merge)"
746 "set ignoreeof"
747 "set fignore =(.obj .pdb .bsc .ilk .idb  .OBJ .PDB .BSC .ILK .IDB)"
748 "set color=(ls-F)"
749 "alias	ls  ls-F "
750 "alias 	h 	history"
751 "alias 	rm 	c:/bin/rm.exe -i"
752 "alias 	rd	c:/bin/rmdir.exe"
753 "alias 	md	c:/bin/mkdir.exe"
754 "alias 	mv	C:/bin/mv.exe -i"
755 "alias   cp  C:/bin/cp.exe -i"
756 "alias	pwd 'echo $cwd'"
757 "alias   env printenv"
758 "alias   unsave unset savehist savedirs"
END
#endif 0
